var ApiConnector = require('./api-connector');
var Config = require('./config');
var fs = require('fs');
var Datapoint = require('./datapoint');
var RWLock = require('rwlock');
var lock = new RWLock();
/**
 * Entry point for starting the BitSensor Web Application Security plugin.
 *
 * @constructor
 * @param {string} config Path to <code>config.json</code>, relative to <code>process.cwd()</code>
 */
var BitSensor = function (config) {

    /**
     * @type {Handler[]}
     */
    this.preHandlers = [];
    this.postHandlers = [];
    this.session = require("continuation-local-storage").createNamespace("session");
    this.isDisabled;
    this.isHandlingRequest;
    this.config = new Config(fs.readFileSync(config, 'utf8'));

};

BitSensor.VERSION = '0.1.0';

BitSensor.prototype.runSession = function (callback) {

    this.session.run(function () {
        bitSensor.initialize();
        callback();
        bitSensor.stopTrace();
    });
};

BitSensor.prototype.apply = function (consumer) {
    if (this.hasBeenDisabled()) {
        return;
    }

    if (!this.isInitialized()) {
        bitSensor.runSession(function () {
            consumer.call(this, bitSensor.getDatapoint());
            return;
        });
    }

    consumer.call(this, bitSensor.getDatapoint());


};

BitSensor.prototype.addPreHandler = function (handler) {
    this.preHandlers.push(handler);
};

BitSensor.prototype.addPostHandler = function (handler) {
    this.postHandlers.push(handler);
};


BitSensor.prototype.runPreHandlers = function () {
    this.preHandlers.forEach(function (handler) {
        handler.handle(bitSensor.getDatapoint());
    });
};

BitSensor.prototype.runPostHandlers = function () {
    this.postHandlers.forEach(function (handler) {
        handler.handle(bitSensor.getDatapoint());
    });
};

BitSensor.prototype.startTrace = function () {
    if (!this.isInitialized()) {
        this.initialize();
    }
    this.traceIncrement();
};

BitSensor.prototype.stopTrace = function () {
    if (!this.isInitialized()) {
        return;
    }

    if (this.traceDecrement() > 0) {
        return;
    }

    this.runPostHandlers();
    //send datapoint away.
    this.log(function (response) {
        //check response?
    });
};

BitSensor.prototype.log = function (handler) {
    var config = this.config;

    new ApiConnector().from(this.config.user, this.config.apiKey)
        .to(this.config.uri)
        .with(bitSensor.getDatapoint().toArray())
        .post(ApiConnector.ACTION_LOG)
        .send();
};

BitSensor.prototype.getDatapoint = function () {
    if (this.isInitialized()) {
        var datapoint = this.session.get("datapoint");
        if (datapoint == null) {
            this.session.set("datapoint", createDatapoint());
            return bitSensor.getDatapoint();
        }
        return datapoint;
    }
    return this.createDatapoint();
};

BitSensor.prototype.setDatapoint = function (datapoint) {
    if (datapoint == null) {
        datapoint = this.createDatapoint();
    }
    this.session.set("datapoint", datapoint);
};

BitSensor.prototype.initialize = function () {
    if (this.isInitialized()) {
        this.startTrace();
        return;
    }
    this.setDatapoint(this.createDatapoint());
    this.initializeEnabled();
    this.runPreHandlers();
    this.startTrace();
};

BitSensor.prototype.isInitialized = function () {
    if (this.session == null) {
        return false;
    }
    if (this.session.get('datapoint') == null) {
        return false;
    }
    return this.session.get('datapoint') != null;
};

BitSensor.prototype.hasBeenDisabled = function () {
    if (bitSensor.isDisabled == null) {
        return false;

    }
    return bitSensor.isDisabled > 0;
};

BitSensor.prototype.setDisabled = function () {
    this.initializeEnabled();
    lock.writeLock(function (release) {
        bitSensor.isDisabled++;
        release();
    });
    return bitSensor.isDisabled;
};

BitSensor.prototype.setEnabled = function () {
    this.initializeEnabled();
    lock.writeLock(function (release) {
        if (bitSensor.isDisabled > 0) {
            bitSensor.isDisabled--;
            //ignore additional disable
        }
        release();
    });
    return bitSensor.isDisabled;
};

BitSensor.prototype.initializeEnabled = function () {
    lock.writeLock(function (release) {
        if (bitSensor.isDisabled == null) {
            bitSensor.isDisabled = 0;
        }
        if (bitSensor.isHandlingRequest == null) {
            bitSensor.isHandlingRequest = 0;
        }
        release();
    });
};

BitSensor.prototype.traceIncrement = function () {
    if (this.isHandlingRequest == null) {
        this.initialize();
    }
    lock.writeLock(function (release) {
        bitSensor.isHandlingRequest++;
        release();
    });
};

BitSensor.prototype.traceDecrement = function () {
    if (this.isHandlingRequest == null) {
        this.initialize();
    }
    lock.writeLock(function (release) {
        if(bitSensor.isHandlingRequest > 0) {
            bitSensor.isHandlingRequest--;
        }
        release();
    });
    return this.isHandlingRequest;
};

BitSensor.prototype.createDatapoint = function () {

    return new Datapoint();
};

module.exports = BitSensor;